#include "ServerController.h"
#include "../lib/cJSON/cJSON.h"
#include <math.h>

#define CONFIG "../config.cfg"
#define SECRETS ".secrets"

const int PORT = 8888;
struct MHD_Daemon *mhdDaemon;
int coreCount;
float *cpuUtilization;
unsigned int *lastProcStat;
unsigned int maxRam;
unsigned int usedRam;
unsigned int maxSwap;
unsigned int usedSwap;
char *uptime;

void sigintHandler(int signal){
	if (mhdDaemon != NULL){
		printf("Stopping daemon...\n");
		MHD_stop_daemon(mhdDaemon);
		printf("Daemon stopped.\n");
	}
	free(cpuUtilization);
	free(lastProcStat);
	printf("Handled sigint\n");
	exit(0);
}

void error(char *msg){
	write(0, msg, strlen(msg));
	exit(1);
}

void initUtilization(){
	uptime = malloc(16);

	FILE *stats = fopen("/proc/stat", "r");
	char *buf = malloc(128);
	unsigned int *procStat = malloc(sizeof(int) * (coreCount + 1) * 4);
	for (int i = 0; i < coreCount + 1; i++){
		fgets(buf, 128, stats);
		printf("%s", buf);
		sscanf(buf, "%*s %i %i %i %i", &procStat[(4 * i)], &procStat[(4 * i) + 1], &procStat[(4 * i) + 2], &procStat[(4 * i) + 3]);
	}
	fclose(stats);
	free(lastProcStat);
	lastProcStat = procStat;

	stats = fopen("/proc/meminfo", "r");
	while (fgets(buf, 128, stats)){
		char *name = malloc(32);
		unsigned int kb = 0;
		sscanf(buf, "%s %u", name, &kb);
		printf("Name: %s   Value: %u\n", name, kb);
		if (strcmp(name, "MemTotal:") == 0){
			maxRam = kb;
		}else if (strcmp(name, "SwapTotal:") == 0){
			maxSwap = kb;
		}
		free(name);
	}

	free(buf);
}

void updateUtilization(){
	struct sysinfo info;
	sysinfo(&info);
	long int time = info.uptime;
	char years = (char) (time / 31557600);
	time = time % 31557600;
	short int days = (short int) (time / 86400);
	time = time % 86400;
	char hours = (char) (time / 3600);
	time = time % 3600;
	char minutes = (char) (time / 60);
	sprintf(uptime, "%i:%i:%i:%i", years, days, hours, minutes);

	FILE *stats = fopen("/proc/stat", "r");
	char *buf = malloc(128);
	unsigned int *procStat = malloc(sizeof(int) * (coreCount + 1) * 4);
	for (int i = 0; i < coreCount + 1; i++){
		fgets(buf, 128, stats);
		sscanf(buf, "%*s %i %i %i %i", &procStat[(4 * i)], &procStat[(4 * i) + 1], &procStat[(4 * i) + 2], &procStat[(4 * i) + 3]);
	}
	for (int i = 0; i < coreCount + 1; i++){
		unsigned int user = procStat[i * 4] - lastProcStat[i * 4];
		unsigned int nice = procStat[(i * 4) + 1] - lastProcStat[(i * 4) + 1];
		unsigned int system = procStat[(i * 4) + 2] - lastProcStat[(i * 4) + 2];
		unsigned int idle = procStat[(i * 4) + 3] - lastProcStat[(i * 4) + 3];
		cpuUtilization[i] = 100.0 * (float) (user + nice + system) / (float) (user + nice + system + idle);
	}
	fclose(stats);
	free(lastProcStat);
	lastProcStat = procStat;
	
	stats = fopen("/proc/meminfo", "r");
	while (fgets(buf, 128, stats)){
		char *name = malloc(32);
		unsigned int kb;
		sscanf(buf, "%s %u", name, &kb);
		if (strcmp(name, "MemAvailable:") == 0){
			usedRam = maxRam - kb;
		}else if (strcmp(name, "SwapFree:") == 0){
			usedSwap = maxSwap - kb;
		}
		free(name);
	}
	fclose(stats);

	free(buf);
}

enum MHD_Result sendPage(struct MHD_Connection *connection, char* page, char *mimetype){
    struct MHD_Response *response = MHD_create_response_from_buffer(strlen(page), (void *) page, MHD_RESPMEM_PERSISTENT);
	MHD_add_response_header(response, "Content-Type", mimetype);
    enum MHD_Result result = MHD_queue_response(connection, MHD_HTTP_OK, response);
    MHD_destroy_response(response);
    return result;
}

int answerConnection(void *cls, struct MHD_Connection *connection, const char *url, const char *method, const char *version, const char *uploadData, size_t *uploadDataSize, void **con_ls){
	if (strcmp(url, "/server") == 0){
		cJSON *response = cJSON_CreateObject();
		cJSON_AddStringToObject(response, "uptime", uptime);
		cJSON_AddNumberToObject(response, "cpuCount", coreCount);
		for (short int i = 0; i < coreCount + 1; i++){
			char *buf = malloc(32);
			sprintf(buf, "cpu%i", i);
			cJSON_AddNumberToObject(response, buf, (int) (cpuUtilization[i] + 0.5));
			free(buf);
		}
		cJSON_AddNumberToObject(response, "totalRam", maxRam);
		cJSON_AddNumberToObject(response, "usedRam", usedRam);
		cJSON_AddNumberToObject(response, "totalSwap", maxSwap);
		cJSON_AddNumberToObject(response, "usedSwap", usedSwap);

		enum MHD_Result result = sendPage(connection, cJSON_Print(response), "application/json");
		
		cJSON_Delete(response);
		return result;
	}
	return MHD_NO;
}

void readConfig(){
	FILE *file = fopen("config.cfg", "r");
	char buf[1024];
	while (fgets(buf, 1024, file) != NULL){
		char *newbuf = calloc(1024, 1);
		for (int i = 0; i < 1024; i++){
			if (buf[i] == '#') break;
			else newbuf[i] = buf[i];
		}
		char buf1[512];
		char buf2[512];
		if (sscanf(newbuf, "%s %s\n", buf1, buf2) == 2){
			printf("(%s,%s)\n", buf1, buf2);
		}
		free(newbuf);
	}
}

int main(){
	signal(SIGINT, sigintHandler);
	readConfig();
	coreCount = sysconf(_SC_NPROCESSORS_ONLN);
	cpuUtilization = malloc(sizeof(float) * (coreCount + 1));
	lastProcStat = malloc(sizeof(int) * (coreCount + 1) * 4);
	lastProcStat[0] = -1;
	initUtilization();
	mhdDaemon = MHD_start_daemon(MHD_USE_INTERNAL_POLLING_THREAD, PORT, NULL, NULL, (MHD_AccessHandlerCallback) &answerConnection, NULL, MHD_OPTION_END);
	if (mhdDaemon == NULL) error("ERROR: Daemon failed to initialize\n");

	while (1){
		usleep(1000000);
		updateUtilization();
		// for (int i = 0; i < coreCount + 1; i++){
		// 	printf("core%i: %f%%\n", i, cpuUtilization[i]);
		// }
		// printf("RAM Usage: %f%%\nSwap Usage: %f%%\n", 100.0 * (float) usedRam / (float) maxRam, 100.0 * (float) usedSwap / (float) maxSwap);
	}

	MHD_stop_daemon(mhdDaemon);
	free(cpuUtilization);
	free(lastProcStat);

	return 0;
}
